<?php

/**
 * Programm-Modul fr die Verwaltung und Anzeige aktueller ffnungszeiten.
 *
 * Zur persistenten Speicherung der ffnungszeiten-Regeln werden die folgenden
 * Datenbank-Tabellen bentigt (MySQL):
 *
 * <pre>CREATE TABLE offen_regel (
 *   id           MEDIUMINT UNSIGNED NOT NULL AUTO_INCREMENT,
 *   von          DATETIME           NOT NULL,
 *   bis          DATETIME           NOT NULL,
 *   offen        TINYINT            NOT NULL DEFAULT '1',
 *   typ_id       CHAR(1)            NOT NULL DEFAULT 'w',
 *   wert         SMALLINT  UNSIGNED          DEFAULT '1',
 *   PRIMARY KEY (id)
 * );
 * CREATE TABLE offen_typ (
 *   id           CHAR(1)            NOT NULL,
 *   beschreibung VARCHAR(40)        NOT NULL,
 *   reihenfolge  TINYINT            NOT NULL,
 *   PRIMARY KEY (id)
 * );
 * INSERT INTO offen_typ VALUES
 *   ('1', 'Einmalig',                    1),
 *   ('t', 'Aller x Tag(e)',              2),
 *   ('w', 'Aller x Woche(n)',            3),
 *   ('m', 'Aller x Monat(e)',            4),
 *   ('x', 'Jeden x. Wochentag im Monat', 5),
 *   ('j', 'Aller x Jahr(e)',             6);</pre>
 *
 * @author Thiemo Mttig (http://maettig.com/)
 * @version 2005-02-21
 * @requires PHP 4.0.3 (mysql_fetch_assoc)
 * @package zahnarzt-frankfurt.com
 */
class Oeffnungszeiten
{
    /**
     * Voreingestellt ist "offen_".
     *
     * @var string Prfix der Datenbank-Tabellennamen.
     */
    var $prefix = 'offen_';

    /**
     * @var array Meldungstexte fr die Methode {@link getNaechsteOeffnung}.
     */
    var $meldungen = array(
        'heute' =>
            'Unsere Praxis ist heute {zeitraum} Uhr geffnet.',
        'morgen' =>
            'Unsere Praxis ist heute leider schon geschlossen. Wir sind morgen {zeitraum} Uhr wieder fr Sie da.',
        'spaeter' =>
            'Unsere Praxis ist heute leider schon geschlossen. Wir sind am {wochentag}, den {datum} {zeitraum} Uhr wieder fr Sie da.',
        'nie' =>
            'Unsere Praxis ist vorbergehend nicht besetzt.');

    /**
     * @var array Deutsche Wochentage, da man sich auf strftime("%A") nicht
     * verlasen kann.
     * @access private
     */
    var $wochentage = array('Sonntag', 'Montag', 'Dienstag', 'Mittwoch',
        'Donnerstag', 'Freitag', 'Samstag');

    /**
     * @var array
     * @access private
     */
    var $kalender = array();

    /**
     * Instantiiert ein neues Oeffnungszeiten-Objekt.
     *
     * @return Oeffnungszeiten
     */
    function Oeffnungszeiten()
    {
    }

    /**
     * Gibt alle in der Datenbank definierten Regeltypen zurck.
     *
     * @return array
     */
    function getTypen()
    {
        $typen = array();

        $sql = "SELECT id, beschreibung
                  FROM " . $this->prefix . "typ
                 ORDER BY reihenfolge;";
        $ergebnis = mysql_query($sql);
        while ($typ = mysql_fetch_assoc($ergebnis)) $typen[] = $typ;

        return $typen;
    }

    /**
     * Gibt alle in der Datenbank gespeicherten Regeln zurck.
     *
     * Beim Aufruf dieser Methode (d.h. beim Aufruf der
     * Administrations-Oberflche) werden alle abgelaufenen einmaligen Regeln
     * gelscht.
     *
     * @return array
     */
    function getRegeln()
    {
        // Regeln, die um 23 Uhr anfangen, gelten noch bis zu einen Tag lang.
        $sql = "DELETE FROM " . $this->prefix . "regel
                 WHERE typ_id = '1'
                   AND TO_DAYS(von) < TO_DAYS(CURRENT_DATE) - 1;";
        $ergebnis = mysql_query($sql);

        $regeln = array();

        $sql = "SELECT r.*, t.beschreibung,
                       DATE_FORMAT(r.von, '%Y-%m-%d') as datum,
                       DATE_FORMAT(r.von, '%w') as wochentag,
                       TIME_FORMAT(r.von, '%H:%i') as von,
                       TIME_FORMAT(r.bis, '%H:%i') as bis
                  FROM " . $this->prefix . "regel AS r
                  LEFT OUTER JOIN " . $this->prefix . "typ AS t
                    ON r.typ_id = t.id
                 ORDER BY r.offen DESC, t.reihenfolge, r.von, r.bis, r.wert;";
        if (! $ergebnis = mysql_query($sql)) return array();
        while ($regel = mysql_fetch_assoc($ergebnis))
        {
            $regel['wochentag'] = $this->wochentage[$regel['wochentag']];
            if ($regel['typ_id'] == '1') $regel['wert'] = '';
            if (preg_match('/\bx\b/', $regel['beschreibung']))
            {
                $regel['beschreibung'] = preg_replace('/\bx\b/',
                    $regel['wert'], $regel['beschreibung']);
            }
            elseif ($regel['wert'])
            {
                $regel['beschreibung'] .= ' (' . $regel['wert'] . ')';
            }
            $regeln[] = $regel;
        }

        return $regeln;
    }

    /**
     * Legt eine neue Regel in der Datenbank an oder aktualisiert eine
     * bestehende.
     *
     * @param regel array
     * @return void
     */
    function setRegel($regel)
    {
        // Deutsches Datumsformat in die von MySQL verstandene ISO-Form wandeln.
        if (preg_match('/(\d+)\.\s*(\d+)\.\s*(\d+)/', $regel['datum'], $matches))
        {
            $regel['datum'] = sprinf("%04d-%02d-%02d", $matches[3], $matches[2], $matches[1]);
        }
        foreach (array('von', 'bis') as $key)
        {
            // Erlaubt sind nur Zeiten von "00:00" bis "23:59".
            if (preg_match('/(\d{1,2})\D*(\d*)/', $regel[$key], $matches))
            {
                $regel[$key] = sprintf("%02d:%02d", $matches[1] % 24, $matches[2] % 60);
            }
        }
        if ($regel['typ_id'] == '1') $regel['wert'] = 'NULL';
        else $regel['wert'] = max(1, abs((int)$regel['wert']));

        $sql = "SELECT id
                  FROM " . $this->prefix . "regel
                 WHERE von    = '" . $regel['datum'] . " " . $regel['von'] . "'
                   AND bis    = '" . $regel['datum'] . " " . $regel['bis'] . "'
                   AND offen  = '" . (bool)$regel['offen'] . "'
                   AND typ_id = '" . $regel['typ_id'] . "'
                   AND wert   " . ($regel['typ_id'] == '1' ? "IS" : "=") . " " . $regel['wert'] . ";";
        $ergebnis = mysql_query($sql);
        if (mysql_num_rows($ergebnis) > 0) return;

        if (empty($regel['id']))
        {
            $sql = "INSERT INTO " . $this->prefix . "regel
                           (von, bis, offen, typ_id, wert)
                    VALUES ('" . $regel['datum'] . " " . $regel['von'] . "',
                            '" . $regel['datum'] . " " . $regel['bis'] . "',
                            '" . (bool)$regel['offen'] . "',
                            '" . $regel['typ_id'] . "',
                            " . $regel['wert'] . ");";
        }
        else
        {
            $sql = "UPDATE " . $this->prefix . "regel
                       SET von    = '" . $regel['datum'] . " " . $regel['von'] . "',
                           bis    = '" . $regel['datum'] . " " . $regel['bis'] . "',
                           offen  = '" . (bool)$regel['offen'] . "',
                           typ_id = '" . $regel['typ_id'] . "',
                           wert   = " . $regel['wert'] . "
                     WHERE id     = " . (int)$regel['id'] . ";";
        }
        mysql_query($sql);
    }

    /**
     * Lscht eine Regel aus der Datenbank.
     *
     * @param id int
     * @return void
     */
    function removeRegel($id)
    {
        $sql = "DELETE FROM " . $this->prefix . "regel
                 WHERE id = " . (int)$id . ";";
        mysql_query($sql);
    }

    /**
     * Gibt einen Kalender mit allen ffnungszeiten zurck.
     *
     * Erzeugt eine lckenlose Vorschau auf die ffnungszeiten der nchsten
     * Tage. Voreingestellt sind 28 Tage (4 Wochen). Zeiten, die ber
     * Mitternacht reichen und mehrere Tage berhren, werden fr jeden Tag
     * gesondert angezeigt.
     *
     * @param tage int
     * @return array
     */
    function getKalender($tage = 28)
    {
        $this->_calculate($tage);

        $liste = array();
        for ($i = 0; $i < $tage; ++$i)
        {
            $timestamp = mktime(12, 0, 0, date('n'), date('j') + $i, date('Y'));
            $liste[date('j.n.Y', $timestamp)] = array(
                'wochentag' => $this->wochentage[date('w', $timestamp)],
                'beschreibung' => 'Geschlossen');
        }
        foreach ($this->kalender as $von => $bis)
        {
            while ($von < $bis)
            {
                $eod = mktime(0, 0, 0, date('n', $von), date('j', $von) + 1,
                    date('Y', $von));
                $datum = date('j.n.Y', $von);
                if (isset($liste[$datum]))
                {
                    if ($liste[$datum]['beschreibung'] == 'Geschlossen')
                    {
                        $liste[$datum]['beschreibung'] = '';
                    }
                    else
                    {
                        $liste[$datum]['beschreibung'] .= ', ';
                    }
                    if ($bis >= $eod) $anzeigeBis = '24:00';
                    else $anzeigeBis = date('H:i', $bis);
                    $liste[$datum]['beschreibung'] .= date('H:i', $von) . '-' . $anzeigeBis;
                }
                $von = $eod;
            }
        }
        return $liste;
    }

    /**
     * Gibt die nchste relevante ffnungszeit als Text zurck.
     *
     * Die Berechnung umfasst ausschlielich den angegebenen Zeitraum in Tagen
     * (plus ca. 1 Tag Toleranz). Voreingestellt sind 28 Tage (4 Wochen). Wird
     * in diesem Zeitraum keine ffnungszeit gefunden, erscheint die
     * {@link meldungen Meldung} 'nie'.
     *
     * @param tage int
     * @return string
     */
    function getNaechsteOeffnung($tage = 28)
    {
        $this->_calculate($tage);
        foreach ($this->kalender as $von => $bis)
        {
            if ($bis <= time()) continue;
            if (isset($vonVoriges) && $von >= $vonVoriges) break;
            $vonVoriges = mktime(24, 0, 0, date('n', $von), date('j', $von),
                date('Y', $von));

            $datum = $von;
            if (empty($zeitraum)) $zeitraum = "";
            else $zeitraum .= ' und ';
            $zeitraum .= date('\v\o\n G:i \b\i\s ', $von);
            if ($bis == $vonVoriges)
            {
                $zeitraum .= '24:00';
            }
            elseif ($bis > $vonVoriges)
            {
                $zeitraum .= date(' \z\u\m j.n. \u\m G:i', $bis);
            }
            else
            {
                $zeitraum .= date('G:i', $bis);
            }
        }
        if (empty($datum))
        {
            $meldung = $this->meldungen['nie'];
        }
        elseif ($datum < mktime(24, 0, 0, date('n'), date('j'), date('Y')))
        {
            $meldung = str_replace('{zeitraum}', $zeitraum,
                $this->meldungen['heute']);
        }
        elseif ($datum < mktime(24, 0, 0, date('n'), date('j') + 1, date('Y')))
        {
            $meldung = str_replace('{zeitraum}', $zeitraum,
                $this->meldungen['morgen']);
        }
        else
        {
            $meldung = str_replace('{zeitraum}', $zeitraum,
                $this->meldungen['spaeter']);
            $meldung = str_replace('{datum}',date('j.n.', $datum), $meldung);
            $meldung = str_replace('{wochentag}',
                $this->wochentage[date('w', $datum)], $meldung);
        }
        return $meldung;
    }

    /**
     * @param tage int
     * @return void
     * @access private
     */
    function _calculate($tage)
    {
        $this->kalender = array();

        $sql = "SELECT id, offen, typ_id, wert,
                       TO_DAYS(von) AS von_days,
                       TO_DAYS(CURRENT_DATE) AS current_days,
                       UNIX_TIMESTAMP(von) as von,
                       UNIX_TIMESTAMP(bis) as bis
                  FROM " . $this->prefix . "regel
                 ORDER BY offen DESC;";
        if (! $ergebnis = mysql_query($sql)) return;
        while ($regel = mysql_fetch_assoc($ergebnis))
        {
            if ($regel['typ_id'] == 'w')
            {
                $regel['typ_id'] = 't';
                $regel['wert'] *= 7;
            }

            if ($regel['typ_id'] == '1')
            {
                $this->_set($regel['von'], $regel['bis'], $regel['offen']);
            }
            elseif ($regel['typ_id'] == 't')
            {
                $regelVergleich = $regel['current_days'] - $regel['von_days'];
                for ($i = $tage; $i >= -1; --$i)
                {
                    if (abs($regelVergleich + $i) % $regel['wert']) continue;
                    $this->_set(mktime(date('G', $regel['von']),
                        date('i', $regel['von']), date('s', $regel['von']),
                        date('n'), date('j') + $i, date('Y')),
                        $regel['bis'], $regel['offen']);
                }
            }
            elseif ($regel['typ_id'] == 'm')
            {
                $regelVergleich = date('Y') * 12 + date('n') -
                    date('Y', $regel['von']) * 12 + date('n', $regel['von']);
                for ($i = ceil($tage / 30); $i >= 0; --$i)
                {
                    if (abs($regelVergleich + $i) % $regel['wert']) continue;
                    $this->_set(mktime(date('G', $regel['von']),
                        date('i', $regel['von']), date('s', $regel['von']),
                        date('n') + $i, date('j', $regel['von']), date('Y')),
                        $regel['bis'], $regel['offen']);
                }
            }
            elseif ($regel['typ_id'] == 'x')
            {
                $regelWochentag = date('w', $regel['von']);
                $jahr = date('Y');
                $monat = date('n');
                $tag = date('j');
                $stunde = date('G', $regel['von']);
                $minute = date('i', $regel['von']);
                $sekunde = date('s', $regel['von']);
                for ($i = -28; $i <= $tage; ++$i)
                {
                    $von = mktime($stunde, $minute, $sekunde, $monat, $tag + $i,
                        $jahr);
                    if (date('w', $von) != $regelWochentag) continue;
                    if (date('j', $von) <= 7) $n = 1; else @++$n;
                    if ($n != $regel['wert']) continue;
                    $this->_set($von, $regel['bis'], $regel['offen']);
                }
            }
            elseif ($regel['typ_id'] == 'j')
            {
                $regelVergleich = date('Y') - date('Y', $regel['von']);
                for ($i = 0; $i <= 1; ++$i)
                {
                    if (abs($regelVergleich + $i) % $regel['wert']) continue;
                    $this->_set(mktime(date('G', $regel['von']),
                        date('i', $regel['von']), date('s', $regel['von']),
                        date('n', $regel['von']), date('j', $regel['von']),
                        date('Y') + $i),
                        $regel['bis'], $regel['offen']);
                }
            }
        }

        ksort($this->kalender);
    }

    /**
     * @param von int
     * @param bis int
     * @param offen bool
     * @return void
     * @access private
     */
    function _set($von, $bis, $offen)
    {
        $bis = mktime(date('G', $bis), date('i', $bis), date('s', $bis),
            date('n', $von), date('j', $von), date('Y', $von));
        if ($bis <= $von)
        {
            $bis = mktime(date('G', $bis), date('i', $bis), date('s', $bis),
                date('n', $bis), date('j', $bis) + 1, date('Y', $bis));
        }

        foreach ($this->kalender as $v => $b)
        {
            if ($v >= $von && $b <= $bis)
            {
                unset($this->kalender[$v]);
                continue;
            }
            if ($von >= $v && $von <= $b)
            {
                $von_merge = $v;
                if (isset($bis_merge)) break;
            }
            if ($bis >= $v && $bis <= $b)
            {
                $bis_merge = $v;
                if (isset($von_merge)) break;
            }
        }

        if ($offen)
        {
            if (isset($von_merge) && isset($bis_merge))
            {
                $this->kalender[$von_merge] = $this->kalender[$bis_merge];
                if ($von_merge != $bis_merge) unset($this->kalender[$bis_merge]);
            }
            elseif (isset($von_merge))
            {
                $this->kalender[$von_merge] = $bis;
            }
            elseif (isset($bis_merge))
            {
                $this->kalender[$von] = $this->kalender[$bis_merge];
                unset($this->kalender[$bis_merge]);
            }
            else
            {
                $this->kalender[$von] = $bis;
            }
        }
        else
        {
            if (isset($von_merge) && isset($bis_merge) && $von_merge == $bis_merge)
            {
                $this->kalender[$bis] = $this->kalender[$bis_merge];
                $this->kalender[$von_merge] = $von;
                return;
            }
            if (isset($von_merge))
            {
                $this->kalender[$von_merge] = $von;
            }
            if (isset($bis_merge))
            {
                $this->kalender[$bis] = $this->kalender[$bis_merge];
                unset($this->kalender[$bis_merge]);
            }
        }
    }
}

?>